import numpy as np
import matplotlib.pyplot as plt
from numpy.linalg import inv
from scipy.stats.distributions import chi2
plt.rcParams.update({'font.size': 14})
def compute_y(x, times:list, rain):
    '''Functional model, q: response due to rain event.
    Inputs:
      x: tuple, list or array of parameters (d, a, r)
      times: array of times
      rain: tuple, list or array describing rain event
             (rain (m), start day, stop day)
    Returns: ndarray of groundwater levels due to rain event
    '''
    h = (x[0]
         + (np.heaviside(times - rain[1], 1)
            *rain[0]*x[2]*(1 - np.exp(-(times - rain[1])/x[1]))
            - np.heaviside(times - rain[2], 1)
              *rain[0]*x[2]*(1 - np.exp(-(times - rain[2])/x[1]))
            )
         )
    return h
d = 0
a = 1
r = 100
test_n_days = 25
test_times = np.arange(1, test_n_days+1, 0.1)
test_h_t = compute_y((d, a, r), test_times, (0.05, 4, 7))
plt.figure(figsize=(10, 6))
plt.plot(test_times, test_h_t,'r-', linewidth=4.0)
plt.xlabel('Time [days]')
plt.ylabel('Water level [m]')
plt.xlim([0, test_n_days])
plt.ylim([0, 5]);
n_days = 25
y = np.genfromtxt('./data/well_levels.csv' , delimiter=" ,")
times = np.arange(1, n_days+1, 1)
plt.figure(figsize=(10, 6))
plt.plot(times, y,'co', mec='black')
plt.xlabel('Time [days]')
plt.ylabel('Waterlevel [m]');
def jacobian(x, times, rain):
    '''Compute Jacobian of the functional model.
    Input:
      x: tuple, list or array of parameters (d, a, r)
      times: array of times
      rain: tuple, list or array describing rain event
             (p [m], start [day], stop [day])
    Outputs: The Jacobian
             (partial derivatives w.r.t. d, a, and r)
    '''
    dqdd = np.ones(len(times))
    dqda = (rain[0]*x[2]*( (times - rain[2])
                       *np.exp(-(times - rain[2])/x[1])/x[1]**2
                       *np.heaviside(times - rain[2], 1)
                       -  (times - rain[1])
                          *np.exp(-(times - rain[1])/x[1])/x[1]**2
                          *np.heaviside(times - rain[1], 1)
                      )
            )
    dqdr = (rain[0]*( (1 - np.exp(-(times - rain[1])/x[1]))
                   *np.heaviside(times - rain[1], 1)
                   +  (np.exp(-(times - rain[2])/x[1]) - 1)
                      *np.heaviside(times - rain[2], 1)
                  )
            )
    J = np.column_stack((dqdd, dqda, dqdr))
    return J
d_init = y[0]
a_init = 1
r_init = 3
rain_event = (0.05, 4, 7)
sigma = 0.01
var_Y = sigma**2
inv_Sigma_Y = 1/var_Y * np.eye(len(y))
max_iter = 50
x_norm = 10000  # initialize stop criteria (norm of x)
param_init = np.array([d_init, a_init, r_init])
x_hat_i = np.zeros((3, max_iter))
x_hat_i[:] = np.nan
x_hat_i[:, 0] = param_init
iteration = 0
while x_norm >= 1e-12 and iteration < max_iter - 1:
    y_comp_i = compute_y(x_hat_i[:, iteration], times, rain_event)
    Delta_y_i = y - y_comp_i
    J_i = jacobian(x_hat_i[:, iteration], times, rain_event)
    N_i = J_i.T @ inv_Sigma_Y @ J_i
    Delta_x_hat_i = np.linalg.inv(N_i) @ J_i.T @ inv_Sigma_Y @ Delta_y_i
    x_hat_i[:, iteration+1] = x_hat_i[:, iteration] + Delta_x_hat_i
    x_norm = Delta_x_hat_i.T @ N_i @ Delta_x_hat_i
    iteration += 1
    if iteration == max_iter - 1:
        print("Number of iterations too large, check initial values.")
print('Initial estimates:')
print(f'base level [m]:\t\t {round(x_hat_i[0, 0], 4)}')
print(f'scaling parameter:\t {round(x_hat_i[1, 0], 4)}')
print(f'response [m/m]:\t\t {round(x_hat_i[2, 0], 4)}','\n')
print('Final estimates:')
print(f'base level [m]:\t\t {round(x_hat_i[0, iteration], 4)}')
print(f'scaling parameter:\t {round(x_hat_i[1, iteration], 4)}')
print(f'response [m/m]:\t\t {round(x_hat_i[2, iteration], 4)}')
n_par = np.shape(J_i)[1]
print(f'\nNumber of unknowns:\t {n_par}')
print(f'Number of observations:\t {n_days}')
print(f'Redundancy:\t\t {n_days - n_par}')
y_hat = compute_y((x_hat_i[0, iteration],
                   x_hat_i[1, iteration],
                   x_hat_i[2, iteration]
                   ),
                  times, rain_event)
plt.figure(figsize=(10, 6))
t = np.arange(1, n_days+1, 0.1)
plt.plot(times, y_hat , linewidth=4,
         label='Gauss Newton fit', color='black')
plt.plot(times, y, 'co', mec='black',
         markersize=10, label='Observations')
plt.legend()
plt.xlabel('Time [days]')
plt.ylabel('Water level [m]');
params = ['d', 'a', 'r']
fig, ax = plt.subplots(1,3, figsize=(16,4))
plt.subplots_adjust(wspace=0.35)
for i in range(3):
    ax[i].plot(x_hat_i[i, :].T, linewidth=4)
    ax[i].set_title(f'Convergence of {params[i]}')
    ax[i].set_xlabel(f'Number of iterations')
    ax[i].set_ylabel(f'{params[i]}')
e_hat = y - y_hat
Tq = e_hat.T @ inv_Sigma_Y @ e_hat
alpha = 0.05
q = n_days - n_par
k = chi2.ppf(1 - alpha, q)
if Tq < k:
    print(f"(T = {Tq:.1f}) < (K = {k:.1f}), OMT is accepted.")
else:
    print(f"(T = {Tq:.1f}) > (K = {k:.1f}), OMT is rejected.")
